package com.gic.download.utils;

import java.io.*;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.annotation.PostConstruct;

import com.gic.download.dto.DownloadReportTempDTO;
import org.apache.commons.lang.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import com.alibaba.fastjson.JSON;
import com.gic.api.base.commons.ServiceResponse;
import com.gic.commons.util.EntityUtil;
import com.gic.enterprise.dto.DownloadReportDTO;
import com.gic.enterprise.exception.CommonException;
import com.gic.enterprise.service.DownloadReportApiService;
import com.gic.thirdparty.BucketNameEnum;
import com.gic.thirdparty.FileOperateUtils;
/**
 * 下载
 * @ClassName: DataDownloadUtils 
 * @Description:  
 * @author guojuxing 
 * @date 2020/9/15 10:43 AM 
 */
@Component
public class DataDownloadUtils {
    private static final Logger LOGGER = LogManager.getLogger(DataDownloadUtils.class);

    private static DataDownloadUtils dataDownloadUtils;

    @Autowired
    private DownloadReportApiService downloadReportApiService;

    private static Pattern NUMBER_PATTERN = Pattern.compile("[0-9]*");

    @PostConstruct
    public void init() {
        dataDownloadUtils = this;
        dataDownloadUtils.downloadReportApiService = downloadReportApiService;
    }

    /**
     * 初始化参数
     */
    private DataDownload params;
    private File csvFile = null;
    private BufferedWriter writer = null;
    /**
     * 是否已经写入标题
     */
    private boolean hasWriteTitle = false;
    /**
     * 分页数索引
     */
    private int index = 0;
    /**
     * 腾讯云文件路径
     */
    private String fileUrl;

    /**
     * 初始化生成报表中心记录
     * @Title: createDownloadReport 
     * @Description:
      * @author guojuxing
     * @param dto 
     * @return  java.lang.Integer 
      */
    public static Integer createDownloadReport(DownloadReport dto) {
        DownloadReportDTO record = EntityUtil.changeEntityNew(DownloadReportDTO.class, dto);
        LOGGER.info("服务对象是否为空:{}", dataDownloadUtils.downloadReportApiService == null);
        ServiceResponse<Integer> result = dataDownloadUtils.downloadReportApiService.saveDownloadReport(record);
        if (result.isSuccess()) {
            return result.getResult();
        }
        throw new CommonException(result);
    }

    public DataDownloadUtils() {

    }

    /**
     * 根据ID获取记录，获取现在数据类型和总数信息
     * @param reportId
     * @return
     */
    public static DownloadReportTempDTO getByReportId(Integer reportId) {
        ServiceResponse<DownloadReportDTO> reportResult = dataDownloadUtils.downloadReportApiService
                .getDownloadReport(reportId);
        if (reportResult.isSuccess()) {
            DownloadReportDTO report = reportResult.getResult();
            DownloadReportTempDTO temp = new DownloadReportTempDTO();
            temp.setDataCount(report.getDataCount());
            temp.setDataType(report.getDataType());
            return temp;
        }
        LOGGER.info("根据reportId获取数据错误:{}", JSON.toJSONString(reportResult));
        return null;
    }

    /**
     * 初始化下载工具类
     * @param initParams
     * @throws IOException
     */
    public DataDownloadUtils(DataDownload initParams) throws IOException {
        LOGGER.info("初始化下载工具类参数值:{}", JSON.toJSONString(initParams));
        //初始化字段值
        params = EntityUtil.changeEntityNew(DataDownload.class, initParams);
        //重置临时文件路径
        DataDownloadUtils.mkDir(new File(params.getTempPath()));
        File file = new File(params.getTempPath());
        //初始化数据
        this.csvFile = File.createTempFile(initParams.getFileName(), ".csv", file);
        this.writer = new BufferedWriter((new OutputStreamWriter(new FileOutputStream(csvFile), "gbk")));
    }

    /**
     * 写数据
     * @param dataList
     */
    public void write(List<?> dataList) {
        index++;
        this.writeTitle();
        this.writeData(dataList);
        LOGGER.info("index:" + index + ",totalPage:" + params.getTotalPage());
        if (index == params.getTotalPage()) {
            this.uploadFileToCloud();
            DataDownloadUtils.uploadFile(params.getReportId(), this.fileUrl);
        }
    }

    /**
     * 删除文件夹
     * @param file
     */
    public static void mkDir(File file) {
        if (file.getParentFile().exists()) {
            file.mkdir();
        } else {
            mkDir(file.getParentFile());
            file.mkdir();
        }
    }

    /**
     * csv写入标题
     */
    private void writeTitle() {
        try {
            if (!hasWriteTitle) {
                if (params.getTitleList() != null) {
                    for (int i = 0, length = params.getTitleList().size(); i < length; i++) {
                        writer.append(params.getTitleList().get(i) + ",");
                    }
                    writer.newLine();
                    hasWriteTitle = true;
                }
            }
        } catch (IOException e) {
            LOGGER.warn(e);
        }
    }

    private void uploadFileToCloud() {
        try {
            writer.flush();
            String fieldCode = "/" + System.currentTimeMillis() + "_" + params.getFileName() + ".csv";
            //文件存进腾讯云
            String url = FileOperateUtils.simpleUploadFileFromLocal(csvFile, fieldCode,
                    BucketNameEnum.REPORT_50000.getName());
            if (StringUtils.isNotBlank(url)) {
                url = "https://" + url;
            }
            this.fileUrl = url;
            LOGGER.info("上传文件到腾讯云，路径：" + url);
            csvFile.deleteOnExit();
        } catch (IOException e) {
            LOGGER.warn(e);
        } finally {
            this.close();
        }
    }

    /**
     * 关闭流
     */
    private void close() {
        try {
            if (writer != null) {
                writer.close();
            }
        } catch (IOException e) {
            LOGGER.warn(e);
        }
    }

    /**
     * csv写入数据
     * @param dataList
     */
    private void writeData(List<?> dataList) {
        if (dataList == null || params.getFieldList() == null) {
            return;
        }
        try {
            for (int i = 0; i < dataList.size(); i++) {
                StringBuffer sb = new StringBuffer();
                for (int j = 0, length = params.getFieldList().size(); j < length; j++) {
                    String fieldName = params.getFieldList().get(j);
                    Object o = DataDownloadUtils.getFieldValue(dataList.get(i), fieldName);
                    if (o == null) {
                        sb.append(",");
                    } else if (o instanceof Double) {
                        DecimalFormat format = new DecimalFormat("0.00");
                        sb.append(format.format(o) + ",");
                    } else if (o instanceof Integer) {
                        sb.append(o + ",");
                    } else if (o instanceof Date) {
                        sb.append(new SimpleDateFormat("yyyy-MM-dd HH:mm:ss").format(o) + "\t,");
                    } else {
                        Matcher isNum = NUMBER_PATTERN.matcher(o.toString());
                        if (isNum.matches()) {
                            if (o.toString().length() >= 12) {
                                sb.append(o + "\t,");
                            } else {
                                sb.append(o + ",");
                            }
                        } else {
                            //去除换行符
                            o = o.toString().replace("\n", "");
                            o = o.toString().replace("\r", "");
                            //英文逗号转义
                            o = DataDownloadUtils.trimChars(o.toString());
                            sb.append(o + ",");
                        }
                    }
                }
                writer.append(sb.toString());
                writer.flush();
                writer.newLine();
            }
        } catch (IOException e) {
            LOGGER.warn(e);
        }
    }

    /**
     * @Title: trimRubbishChar
     * @Description: 导出的时候需要对一格的内容进行检查,看是否有非法字符,以免串行
     * @Since: 2016年8月2日 下午8:13:42
     * @Author: niewj
     * @param content
     * @return
     * @throws
     */
    private static String trimChars(String content) {
        if (content == null) {
            return "";
        }
        // @====> 1.包含,同时包含"，那么先替换双引号"为两个"" + 然后在包裹在双引号里。
        if (content.contains(",") && content.contains("\"")) {
            //content = content.replaceAll(",", " "); // 逗号替换为空格
            content = content.replaceAll("\"", "\"\"");
            content = "\"" + content + "\"";
        } else if (content.contains(",") && !content.contains("\"")) {
            // @====> 2.如果仅仅包含逗号,则用引号包裹即可。
            //content = content.replaceAll(",", " "); // 逗号替换为空格
            content = "\"" + content + "\"";
        } else if (content.contains("\"") && !content.contains(",")) {
            // @====> 3.如果仅仅包含引号,则双引号代替一个引号，再最外层在包裹即可。
            content = "\"" + content.replaceAll("\"", "\"\"") + "\"";
        }
        return content;
    }

    /**
     * 直接读取对象属性值, 无视private/protected修饰符, 不经过getter函数.
     */
    private static Object getFieldValue(final Object obj, final String fieldName) {
        Field field = getAccessibleField(obj, fieldName);
        if (field == null) {
            throw new IllegalArgumentException("no field [" + fieldName + "]");
        }
        Object result = null;
        try {
            Method method = obj.getClass()
                    .getMethod("get" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1), null);
            return method.invoke(obj, null);
        } catch (Exception e) {
            LOGGER.warn(e);
        }
        return result;
    }

    private static Field getAccessibleField(final Object obj, final String fieldName) {
        for (Class<?> superClass = obj.getClass(); superClass != Object.class; superClass = superClass
                .getSuperclass()) {
            try {
                Field field = superClass.getDeclaredField(fieldName);
                field.setAccessible(true);
                return field;
            } catch (NoSuchFieldException e) {
                LOGGER.warn(e);
            }
        }
        return null;
    }

    /**
     * 更新报表记录的腾讯云文件url
     * @param reportId
     * @param fileUrl
     */
    public static void uploadFile(Integer reportId, String fileUrl) {
        ServiceResponse<Integer> result = dataDownloadUtils.downloadReportApiService.updateDownloadUrl(reportId,
                fileUrl);
        if (!result.isSuccess()) {
            throw new CommonException(result);
        }
    }

}
